/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  |           =====>>>>> 2002 <<<<=====             |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (input.C)
  
  Written by: Damian J. J. Farnell

  Date:       Wed Jul 17 12:03:05 BST 2002

  Purpose:    Inputs the CCM script data

  ===================================================

*/

#include<iostream.h>
#include<stdlib.h>
#include<math.h>
#include<stdio.h>
#include<string.h>

#include "../include/Constants.h"
#include "../include/Externals.h"
#include "../include/FunctionPrototypes.h"
#include "../include/nr.h"
#include "../include/nrutil.h"

#if (PARALLEL_FLAG>0)
#include<mpi.h>
#endif

FILE *script;

void convert_numeral_to_char(int number, char *character);

// function prototypes.

void  moveAlongHome(void);
double readInFloat(void);
int   readInInteger(void);
void  setArray(int **typeArray);
void  definePointsInBT(void);
void defineGeneralisedPositions(void);

void input(int argc, char *argv[]) 
{
  
  int i, j ,k, l, m, bonds;
  int run_option;
  
  char root_name[100],append[]=".dat";
  char path[]="../Data/";
  char tempPath[100];
  // char path[100];
  char input_filename[100], input[]=".input";
  char temp[100];
  char configurations2[100], configs_name[]=".configs";

  int dummyInt,dummyInt2;
  float dummyFloat;

  char dummy;

  int typeArrayMax[25]={0,5,2,9,2,9,3,3,1,19,1,3,4,1,0,0};
  int **typeArray;


  typeArrayMax[1]=5;typeArrayMax[2]=2;typeArrayMax[3]=9;
  typeArrayMax[4]=2;typeArrayMax[5]=9;typeArrayMax[6]=3;
  typeArrayMax[7]=3;typeArrayMax[8]=1;typeArrayMax[9]=19;

  typeArrayMax[10]=1;typeArrayMax[11]=3;typeArrayMax[12]=4;
  typeArrayMax[13]=1;typeArrayMax[14]=0;typeArrayMax[15]=0;

  typeArray=imatrix(0,25,0,25);

  setArray(typeArray);

  Control.file=1;    // tells program to save CCM equations to file.

  if (argc == 3) {

    strcpy(input_filename,path);
    strcat(input_filename,argv[1]);
    run_option=atoi(argv[2]);
    if (myid==0)
      printf("myid=%d name=%s runoption=%d\n",myid,input_filename,run_option);

  } else {
    
    printf("Input the 1) root CCM path and 2) name of the CCM script file!"
	   "\nExiting!!\n\n");
    exit(0);
    
  }
  
  if (myid==0)
  printf("/-------------------------------------------------\\\n"
	 "|                                                 |\n"
	 "|                                                 |\n"
	 "|     /=====\\     /=====\\     |\\          /|      |\n"
	 "|    /           /            | \\        / |      |\n"
	 "|    |           |            |  \\      /  |      |\n"
	 "|    |           |            |   \\    /   |      |\n"
	 "|    \\           \\            |    \\  /    |      |\n"
	 "|     \\=====/     \\=====/     |     \\/     |      |\n"
	 "|                                                 |\n"
	 "|                                                 |\n"
	 "|           =====>>>>> 2002 <<<<=====             |\n"
	 "|                                                 |\n"
	 "\\-------------------------------------------------/\n"
	 "\n"
	 "===================================================\n"
	 "Name:       ccm.C\n"
	 "@Copyright: Damian J. J. Farnell\n"
	 "Date:       Wed Jul 17 12:03:05 BST 2002\n"
	 "Purpose:    Determine and solve CCM LSUBm equations.\n"
	 "===================================================\n\n");

  script=fopen(input_filename,"r");
  if (!script) {
    printf("Cannot Open %20s ...\nExiting!!\n\n",input_filename);
    exit(1);
  }

  // read in `control' parameters first:

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "Control Parameters:\n");

  moveAlongHome(); Control.output=readInInteger();   // output level= ...

  if (Control.output > 0 && myid ==0)
    printf("Output (0-non to >6 verbose) =%d\n",Control.output);
 
  moveAlongHome(); fscanf(script,"%80s",tempPath);  // work directory= ...

  moveAlongHome(); fscanf(script,"%80s",temp);      // output file= ...
  
  memcpy(Control.name,tempPath,strlen(tempPath));
#if (PARALLEL_FLAG>0)
  char processorLabel[100]="";
  convert_numeral_to_char(myid,processorLabel);
  strcat(Control.name,"myid_");
  strcat(Control.name,processorLabel);
  strcat(Control.name,"_");
#endif
  strcat(Control.name,temp);

  if (myid==0)
  printf("Simulation Filename = %30s : myid == %d\n",Control.name,myid);

  // read in parameters for solve.C

  
  // now read in the `solve' variables

  moveAlongHome();
  minimum=readInFloat();
  if (Control.output > 0 && myid ==0)
    printf("Minimum delta=%20.10f\n",minimum);

  moveAlongHome();
  maximum=readInFloat();
  if (Control.output > 0 && myid ==0)
    printf("Maximum delta=%20.10f\n",maximum);

  moveAlongHome();
  increment=readInFloat();
  if (Control.output > 0 && myid ==0)
    printf("Increment delta=%30.20f\n",increment);

  // next two variables are minimum and maximum y values
  // used in plot.tcl but not in ccm C program.

  moveAlongHome();
  dummyFloat=readInFloat();
  moveAlongHome();
  dummyFloat=readInFloat();

  // read in highest level parameters first:

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	 "General Parameters:\n");
  
  // dimension

  moveAlongHome();
  Params.dim=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("Dimension=%d\n",Params.dim);

  // spin number

  moveAlongHome();
  Params.s=readInFloat();
  if (Control.output > 0 && myid ==0)
    printf("Maximum spin quantum number=%f\n",Params.s);

  // Bounding Triangle

  dummyInt=0;

  moveAlongHome();
  Params.btLength=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("BT Length=%d\n",Params.btLength);
  
  // SUBn-m: n=

  moveAlongHome();
  Params.n=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("SUBn-m: n=%d\n",Params.n);
  
  // SUBn-m: m=

  moveAlongHome();
  Params.m=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("SUBn-m: m=%d\n",Params.m);
  
  // Number of Bonds: 

  moveAlongHome();
  Params.bonds=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("Bonds: m=%d\n",Params.bonds);
  
  moveAlongHome();
  restrictionClusters=readInInteger();
  if (Control.output > 0 && myid ==0)
    printf("Restriction on Clusters: Type=%d\n",restrictionClusters);
  
  moveAlongHome();
  solutionStrategy=readInInteger();
  if (Control.output > 0 && myid ==0)
    if (solutionStrategy == 1)
      printf("Solve CCM equations using Newton-Raphson method\n");
    else
      printf("Solve CCM equations using Direct Iteration method\n");
  
  if (PARALLEL_FLAG > 0 && solutionStrategy == 1) {
    if (myid==0)
      printf("PARALLEL_FLAG > 0 && solutionStrategy == 1...\n"
	     "Exiting!!\n\n");
    exit(1);
  }

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Unit Cell:\n");
  
  if (Control.output > 1&& myid ==0) 
    printf("Dimension=%d\ns=%5.3f\nm=%d\nB.T.Length=%d\n"
	   "#Sites in B.T.=%d\n\n",
	   Params.dim,
	   Params.s,
	   Params.m,
	   Params.btLength,
	   Params.btLength*(Params.btLength+1)/2);

  if (Params.dim == 1) 
    max_number_cells_BT=Params.btLength;
  else if (Params.dim == 2) 
    max_number_cells_BT=Params.btLength*(Params.btLength+1)/2;
  else {
    max_number_cells_BT=0;
    for(m=0;m<Params.btLength;m++) 
      for (l=0;l<Params.btLength-m;l++) 
	for(k=0;k<Params.btLength-l;k++) 
	  max_number_cells_BT++;
  }

  /*  if ( max_number_cells_BT >= MAX_BT) {
    cout << "Number of Cells in BT is greater than MAX_BT...\n"
	 << "Exiting!\n\n" ;
    exit(1);
    } */

  // probably a good idea to keep 
  // Params.btLength >= Params.m/UnitCell.number+1;

  for(i=0;i<MAX_UNIT_CELL;i++) {
    UnitCell.x[i]=0.0;
    UnitCell.y[i]=0.0;
    UnitCell.z[i]=0.0;
    UnitCell.s[i]=0.5;
  }

  moveAlongHome();
  fscanf(script,"%d",&UnitCell.number);
  if (Control.output > 0 && myid ==0)
    printf("Number of sites in Unit Cell = %d\n",UnitCell.number);
  if (Control.output > 0 && myid ==0)
    printf("Positions & Spin Quantum Number:\n");
  
  for(i=1;i<=UnitCell.number;i++) {
    
    UnitCell.x[i]=readInFloat();
    UnitCell.y[i]=readInFloat();
    UnitCell.z[i]=readInFloat();
    UnitCell.s2[i]=readInFloat();
    UnitCell.s[i]=fabs(UnitCell.s2[i]);
    //    moveAlongHome();

  }

  for(i=1;i<=UnitCell.number;i++) {
    if (Control.output > 0 && myid ==0)
      printf("Site=%d : position = (%f,%f,%f) : spin = %f\n",i,
	     UnitCell.x[i],
	     UnitCell.y[i],
	     UnitCell.z[i],
	     UnitCell.s[i]);

    if (UnitCell.s[i] > Params.s) {
      printf("UnitCell.s[%d] > Params.s...\nExiting!!\n\n",i);
      exit(1);
    }

  }

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Bravais Lattice:\n");

  // define the bravais lattice 

  moveAlongHome();
  BravaisLattice.a[1]=readInFloat();
  BravaisLattice.a[2]=readInFloat();
  BravaisLattice.a[3]=readInFloat();
  BravaisLattice.b[1]=readInFloat();
  BravaisLattice.b[2]=readInFloat();
  BravaisLattice.b[3]=readInFloat();
  BravaisLattice.c[1]=readInFloat();
  BravaisLattice.c[2]=readInFloat();
  BravaisLattice.c[3]=readInFloat();
  
  if (Control.output > 0 && myid ==0) {
    printf("Bravais Lattice Vector A=(%6.4f,%6.4f,%6.4f)\n",
	   BravaisLattice.a[1],BravaisLattice.a[2],BravaisLattice.a[3]);
    printf("Bravais Lattice Vector B=(%6.4f,%6.4f,%6.4f)\n",
	   BravaisLattice.b[1],BravaisLattice.b[2],BravaisLattice.b[3]);
    printf("Bravais Lattice Vector C=(%6.4f,%6.4f,%6.4f)\n",
	   BravaisLattice.c[1],BravaisLattice.c[2],BravaisLattice.c[3]);
  }

  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "Nearest-Neighbours to sites in Unit Cell.:\n");
  

  for(i=1;i<=UnitCell.number;i++) {
    moveAlongHome();
    LatticeNeighbours.number[i]=readInInteger(); 
    for(j=1;j<=LatticeNeighbours.number[i];j++) {
	  LatticeNeighbours.x[i][j]=readInInteger();
	  LatticeNeighbours.y[i][j]=readInInteger();
	  LatticeNeighbours.z[i][j]=readInInteger();
	  LatticeNeighbours.cell[i][j]=readInInteger();
    }
  }

  if (Control.output > 0 && myid ==0) {
    for(i=1;i<=UnitCell.number;i++) {
      printf("Number of nearest-neighbours for site =%d : \n",
	     LatticeNeighbours.number[i]);
      for(j=1;j<=LatticeNeighbours.number[i];j++) {
	printf("#%d : Bravais translation=(%d,%d,%d) to cell (%d)\n",
	       j, LatticeNeighbours.x[i][j],LatticeNeighbours.y[i][j],
	       LatticeNeighbours.z[i][j],LatticeNeighbours.cell[i][j]);
      }
    }
    printf("=======================================\n"
	   "The symmetry operations:\n");
  }
  moveAlongHome();

  LatticeSymmetries.number=readInInteger(); 

  // just to make sure set to sero initially.

  for(i=1;i<=LatticeSymmetries.number;i++){
    LatticeSymmetries.x[i]=0.0;
    LatticeSymmetries.y[i]=0.0;
    LatticeSymmetries.z[i]=0.0;
    for(j=1;j<=3;j++){
      for(k=1;k<=3;k++){
	LatticeSymmetries.R[i][j][k]=0.0;
      }
    }
  }

  for(i=1;i<=LatticeSymmetries.number;i++) {
    for(j=1;j<=3;j++){
      for(k=1;k<=3;k++){
	LatticeSymmetries.R[i][j][k]=readInFloat();
      }
      if (j==1)	
	LatticeSymmetries.x[i]=readInFloat();
      else if (j==2) 
	LatticeSymmetries.y[i]=readInFloat();
      else 
	LatticeSymmetries.z[i]=readInFloat();
    }
  }

  if (Control.output > 0 && myid ==0) {
    for(i=1;i<=LatticeSymmetries.number;i++) {
      printf("Symmetry Operation = %d\nRotation matrix : Prim. Transl. \n",i);
      for(j=1;j<=3;j++){
	for(k=1;k<=3;k++){
	  printf("%5.3f ", LatticeSymmetries.R[i][j][k]);
	}
	if (j==1)	
	  printf(" : %5.3f\n",LatticeSymmetries.x[i]);
	else if (j==2) 
	  printf(" : %5.3f\n",LatticeSymmetries.y[i]);
	else 
	  printf(" : %5.3f\n",LatticeSymmetries.z[i]);
      }
    }
  }

  // now determine the Hamiltonian. The `types' of term
  // first...
  
  if (Control.output > 0 && myid ==0)
    printf("=======================================\n"
	   "The Hamiltonian:\n");
  
  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(i=0;i<MAX_UNIT_CELL;i++) {
      for(j=0;j<MAX_NUM_NEIGHBOURS;j++) {
	for(k=0;k<15;k++) {
	  Hamiltonian_types[bonds].k_to_m_number[i][k]=0;
	  Hamiltonian_types[bonds].m_to_k_number[i][k]=0;
	  Hamiltonian_types[bonds].k_to_m_x[i][j][k]=0;
	  Hamiltonian_types[bonds].k_to_m_y[i][j][k]=0;
	  Hamiltonian_types[bonds].k_to_m_z[i][j][k]=0;
	  Hamiltonian_types[bonds].k_to_m_cell[i][j][k]=0;
	  Hamiltonian_types[bonds].m_to_k_x[i][j][k]=0;
	  Hamiltonian_types[bonds].m_to_k_y[i][j][k]=0;
	  Hamiltonian_types[bonds].m_to_k_z[i][j][k]=0;
	  Hamiltonian_types[bonds].m_to_k_cell[i][j][k]=0;
	  Hamiltonian_types[bonds].coefficient[i][j][k]=0.0;
	}
      }
    }
  }

  moveAlongHome();
  directional=readInInteger();

  if (myid==0)
    printf("Directional?=%d\n",directional);

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(i=1;i<=UnitCell.number;i++) {
      for(k=1;k<14;k++) {

	moveAlongHome();
	Hamiltonian_types[bonds].k_to_m_number[i][k]=readInInteger();
	
	if (k<=9) {

	  if (directional) {
	    for(j=1;j<=Hamiltonian_types[bonds].k_to_m_number[i][k];j++) {
	      Hamiltonian_types[bonds].k_to_m_x[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].k_to_m_y[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].k_to_m_z[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].k_to_m_cell[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].coefficient[i][j][k]=readInFloat();
	    }

	    moveAlongHome();
	    Hamiltonian_types[bonds].m_to_k_number[i][k]=readInInteger();

	    for(j=1;j<=Hamiltonian_types[bonds].m_to_k_number[i][k];j++) {
	      Hamiltonian_types[bonds].m_to_k_x[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].m_to_k_y[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].m_to_k_z[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].m_to_k_cell[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].coefficient[i][j][k]=readInFloat();
	    }

	  } else {

	    Hamiltonian_types[bonds].m_to_k_number[i][k]=
	      Hamiltonian_types[bonds].k_to_m_number[i][k];

	    for(j=1;j<=Hamiltonian_types[bonds].k_to_m_number[i][k];j++) {

	      Hamiltonian_types[bonds].k_to_m_x[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].k_to_m_y[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].k_to_m_z[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].k_to_m_cell[i][j][k]=readInInteger();
	      Hamiltonian_types[bonds].coefficient[i][j][k]=readInFloat();

	      Hamiltonian_types[bonds].m_to_k_x[i][j][k]=
		Hamiltonian_types[bonds].k_to_m_x[i][j][k];
	      Hamiltonian_types[bonds].m_to_k_y[i][j][k]=
		Hamiltonian_types[bonds].k_to_m_y[i][j][k];
	      Hamiltonian_types[bonds].m_to_k_z[i][j][k]=
		Hamiltonian_types[bonds].k_to_m_z[i][j][k];
	      Hamiltonian_types[bonds].m_to_k_cell[i][j][k]=
		Hamiltonian_types[bonds].k_to_m_cell[i][j][k];	      

	    }
	  }
	} else {
	  Hamiltonian_types[bonds].coefficient[i][1][k]=readInFloat();
	}
      }
    }
  }

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(k=1;k<14;k++) {
      Hamiltonian_types[bonds].evaluate[k]=0;
      for(i=1;i<=UnitCell.number;i++) {
	for(j=1;j<=Hamiltonian_types[bonds].k_to_m_number[i][k];j++) {
	  if (fabs(Hamiltonian_types[bonds].coefficient[i][j][k]) > 0.01) {
	    Hamiltonian_types[bonds].evaluate[k]=1;
	    break;
	  }
	}
	if (Hamiltonian_types[bonds].evaluate[k])
	  break;
      }
    }
  }

  // now print out this information:
    
  if (Control.output > 0 && myid ==0) {
    for(bonds=1;bonds<=Params.bonds;bonds++) {
      for(k=1;k<14;k++) {
	if (Control.output>0) 
	  printf("Bond = %d : Type = %d : Evaluate =%d\n"
		 ,bonds,k,
		 Hamiltonian_types[bonds].evaluate[k]);
	if (Control.output>3) {
	  if (k > 9) {
	    for(i=1;i<=UnitCell.number;i++) {
	      printf("Site = %d : # Single Site Term : ",i);
	      printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		     Hamiltonian_types[bonds].k_to_m_x[i][1][k],
		     Hamiltonian_types[bonds].k_to_m_y[i][1][k],
		     Hamiltonian_types[bonds].k_to_m_z[i][1][k],
		     Hamiltonian_types[bonds].k_to_m_cell[i][1][k],
		     Hamiltonian_types[bonds].coefficient[i][1][k]);
	    }
	  } else {

	    for(i=1;i<=UnitCell.number;i++) {

	      if (directional) {
		
		printf("Site = %d : Number Directional Neighbours k to m = %d\n",
		       i,Hamiltonian_types[bonds].k_to_m_number[i][k]);
		
		for(j=1;j<=Hamiltonian_types[bonds].k_to_m_number[i][k];j++) {
		  printf("(%d,%d,%d,%d) : Coeff. =%f\n",
			 Hamiltonian_types[bonds].k_to_m_x[i][j][k],
			 Hamiltonian_types[bonds].k_to_m_y[i][j][k],
			 Hamiltonian_types[bonds].k_to_m_z[i][j][k],
			 Hamiltonian_types[bonds].k_to_m_cell[i][j][k],
			 Hamiltonian_types[bonds].coefficient[i][j][k]);
		}
		printf("Site = %d : Number Directional Neighbours m to k = %d\n",i,
		       Hamiltonian_types[bonds].m_to_k_number[i][k]);
		for(j=1;j<=Hamiltonian_types[bonds].m_to_k_number[i][k];j++) {
		  printf("(%d,%d,%d,%d) : Coeff. =%f\n",
			 Hamiltonian_types[bonds].m_to_k_x[i][j][k],
			 Hamiltonian_types[bonds].m_to_k_y[i][j][k],
			 Hamiltonian_types[bonds].m_to_k_z[i][j][k],
			 Hamiltonian_types[bonds].m_to_k_cell[i][j][k],
			 Hamiltonian_types[bonds].coefficient[i][j][k]);
		}
	      } else {
		
		printf("Site = %d : Number Non-Directional Neighbours  = %d\n"
		       ,i,Hamiltonian_types[bonds].k_to_m_number[i][k]);
		
		for(j=1;j<=Hamiltonian_types[bonds].k_to_m_number[i][k];j++) {
		  printf("(%d,%d,%d,%d) : Coeff. =%f\n",
			 Hamiltonian_types[bonds].k_to_m_x[i][j][k],
			 Hamiltonian_types[bonds].k_to_m_y[i][j][k],
			 Hamiltonian_types[bonds].k_to_m_z[i][j][k],
			 Hamiltonian_types[bonds].k_to_m_cell[i][j][k],
			 Hamiltonian_types[bonds].coefficient[i][j][k]);
		}
	      }
	    }
	  }
	}
      }
    }
  }

  // now input the information for the Hamiltonian proper for each term
  // in the similarity transformed version of H.

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    for(k=1;k<14;k++) {

      if (Hamiltonian_types[bonds].evaluate[k]) {

	for(l=1;l<=typeArrayMax[k];l++) {

	  Hamiltonian[bonds].evaluate[typeArray[k][l]]=1;

	  if (k > 9) {
	    for(i=1;i<=UnitCell.number;i++) {
	      Hamiltonian[bonds].k_to_m_x[i][1][typeArray[k][l]]=
		Hamiltonian_types[bonds].k_to_m_x[i][1][k];
	      Hamiltonian[bonds].k_to_m_y[i][1][typeArray[k][l]]=
		Hamiltonian_types[bonds].k_to_m_y[i][1][k];
	      Hamiltonian[bonds].k_to_m_z[i][1][typeArray[k][l]]=
		Hamiltonian_types[bonds].k_to_m_z[i][1][k];
	      Hamiltonian[bonds].k_to_m_cell[i][1][typeArray[k][l]]=
		Hamiltonian_types[bonds].k_to_m_cell[i][1][k];
	      Hamiltonian[bonds].coefficient[i][1][typeArray[k][l]]=
		Hamiltonian_types[bonds].coefficient[i][1][k];
	    }
	  } else {

	    for(i=1;i<=UnitCell.number;i++) {
	      
	      Hamiltonian[bonds].k_to_m_number[i][typeArray[k][l]]=
		Hamiltonian_types[bonds].k_to_m_number[i][k];
	      
	      for(j=1;j<=Hamiltonian_types[bonds].k_to_m_number[i][k];j++) {

		Hamiltonian[bonds].k_to_m_x[i][j][typeArray[k][l]]=
		  Hamiltonian_types[bonds].k_to_m_x[i][j][k];
		Hamiltonian[bonds].k_to_m_y[i][j][typeArray[k][l]]=
		  Hamiltonian_types[bonds].k_to_m_y[i][j][k];
		Hamiltonian[bonds].k_to_m_z[i][j][typeArray[k][l]]=
		  Hamiltonian_types[bonds].k_to_m_z[i][j][k];
		Hamiltonian[bonds].k_to_m_cell[i][j][typeArray[k][l]]=
		  Hamiltonian_types[bonds].k_to_m_cell[i][j][k];
		Hamiltonian[bonds].coefficient[i][j][typeArray[k][l]]=
		  Hamiltonian_types[bonds].coefficient[i][j][k];

	      }
	      
	      Hamiltonian[bonds].m_to_k_number[i][typeArray[k][l]]=
		Hamiltonian_types[bonds].m_to_k_number[i][k];
	      
	      for(j=1;j<=Hamiltonian_types[bonds].m_to_k_number[i][k];j++) {

		Hamiltonian[bonds].m_to_k_x[i][j][typeArray[k][l]]=
		  Hamiltonian_types[bonds].m_to_k_x[i][j][k];
		Hamiltonian[bonds].m_to_k_y[i][j][typeArray[k][l]]=
		  Hamiltonian_types[bonds].m_to_k_y[i][j][k];
		Hamiltonian[bonds].m_to_k_z[i][j][typeArray[k][l]]=
		  Hamiltonian_types[bonds].m_to_k_z[i][j][k];
		Hamiltonian[bonds].m_to_k_cell[i][j][typeArray[k][l]]=
		  Hamiltonian_types[bonds].m_to_k_cell[i][j][k];
		Hamiltonian[bonds].coefficient[i][j][typeArray[k][l]]=
		  Hamiltonian_types[bonds].coefficient[i][j][k];

	      }
	    }
	  }
	}
      }
    }
  }

  if (Control.output>3 && myid ==0) {

    for(bonds=1;bonds<=Params.bonds;bonds++) {
      printf("\n");

      for(k=1;k<63;k++) {

	printf("Bond = %d : Type = %d : Evaluate =%d\n",bonds,k,
	       Hamiltonian[bonds].evaluate[k]);      

	if (k > 53) {

	  for(i=1;i<=UnitCell.number;i++) {
	    printf("Site = %d : # Single Site Term : ",i);
	    printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		   Hamiltonian[bonds].k_to_m_x[i][1][k],
		   Hamiltonian[bonds].k_to_m_y[i][1][k],
		   Hamiltonian[bonds].k_to_m_z[i][1][k],
		   Hamiltonian[bonds].k_to_m_cell[i][1][k],
		   Hamiltonian[bonds].coefficient[i][1][k]);
	  }

	} else {

	  for(i=1;i<=UnitCell.number;i++) {

	    if (directional) {

	      printf("Site = %d : Number Directional Neighbours k to m = %d\n",
		     i,Hamiltonian[bonds].k_to_m_number[i][k]);
	      for(j=1;j<=Hamiltonian[bonds].k_to_m_number[i][k];j++) {
		printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		       Hamiltonian[bonds].k_to_m_x[i][j][k],
		       Hamiltonian[bonds].k_to_m_y[i][j][k],
		       Hamiltonian[bonds].k_to_m_z[i][j][k],
		       Hamiltonian[bonds].k_to_m_cell[i][j][k],
		       Hamiltonian[bonds].coefficient[i][j][k]);
	      }
	      printf("Site = %d : Number Directional Neighbours m to k = %d\n",i,
		     Hamiltonian[bonds].m_to_k_number[i][k]);
	      for(j=1;j<=Hamiltonian[bonds].m_to_k_number[i][k];j++) {
		printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		       Hamiltonian[bonds].m_to_k_x[i][j][k],
		       Hamiltonian[bonds].m_to_k_y[i][j][k],
		       Hamiltonian[bonds].m_to_k_z[i][j][k],
		       Hamiltonian[bonds].m_to_k_cell[i][j][k],
		       Hamiltonian[bonds].coefficient[i][j][k]);
	      }

	    } else {

	      printf("Site = %d : Number Non-Directional Neighbours  = %d\n"
		     ,i,Hamiltonian[bonds].k_to_m_number[i][k]);

	      for(j=1;j<=Hamiltonian[bonds].k_to_m_number[i][k];j++) {
		printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		       Hamiltonian[bonds].k_to_m_x[i][j][k],
		       Hamiltonian[bonds].k_to_m_y[i][j][k],
		       Hamiltonian[bonds].k_to_m_z[i][j][k],
		       Hamiltonian[bonds].k_to_m_cell[i][j][k],
		       Hamiltonian[bonds].coefficient[i][j][k]);
	      }

	      printf("Site = %d : Number Non-Directional Neighbours  = %d\n"
		     ,i,Hamiltonian[bonds].m_to_k_number[i][k]);

	      for(j=1;j<=Hamiltonian[bonds].k_to_m_number[i][k];j++) {
		printf("(%d,%d,%d,%d) : Coeff. =%f\n",
		       Hamiltonian[bonds].m_to_k_x[i][j][k],
		       Hamiltonian[bonds].m_to_k_y[i][j][k],
		       Hamiltonian[bonds].m_to_k_z[i][j][k],
		       Hamiltonian[bonds].m_to_k_cell[i][j][k],
		       Hamiltonian[bonds].coefficient[i][j][k]);
	      }
	    }
	  }
	}
      }
    }
  }

  fclose(script);
  free_imatrix(typeArray,0,25,0,25);
  
  definePointsInBT();

  //  defineGeneralisedPositions();

  // not used at moment...

  sublatticeMagnetisation.number=UnitCell.number;
  for (i=1;i<=sublatticeMagnetisation.number;i++)
    sublatticeMagnetisation.cell[j]=1;
 
  // read in cluster configurations 
  
  if (run_option == 2 || run_option == 3) {

    strcpy(configurations2,Control.name);
    strcat(configurations2,configs_name);
    fp1=fopen(configurations2,"r");  
    if (!fp1) {
      cout << "Cannot Open " << configurations2 << "...\nExiting!!\n\n";
      exit(1);
    }

    //    cout << configurations2 << endl;

    fread(&NumberConfigurations,sizeof(int),1,fp1);
    
    if (Control.output > 1) {
      printf("\n==========================================================\n"
	     "Fundamental clusters (N_F = %d) are: \n"
	     "============================================================\n\n"
	     ,NumberConfigurations);
    }

    for(i=1;i<=NumberConfigurations;i++) {

      if (Control.output > 1 && myid==0)
	printf("Cluster %d: \n",i);

      fread(spinConfigurations[i].descriptor,
	    ((MAX_SPIN_LIST)*sizeof(int)),
	    1,
	    fp1);
      if (Control.output > 1 && myid==0){
	printOutClusterFromDescriptor(spinConfigurations[i].descriptor);
	printf("\n");
      }
    }
    fclose(fp1);

    /* NEW!!: work out the maximum radius for all clusters 
       This is used to reduce the amount of sweeping for 
       F_k F_m s_k^z s_m^z -like terms. */

    radius=0;
   
    {
       int length,descriptorIndex;
      
      int **tempConfig;
      tempConfig=imatrix(0,MAX_LSUBM+1,0,5);
      
      for(i=1;i<=NumberConfigurations;i++) {
	
	length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
					 tempConfig);
	
	/* note that midPointX, midPointY, midPointZ indicate Unit cells */
	
	// note that integer rounding shouldn't affect this 
	// algorithm
	
	// now find the maximum `radius'
	
	for(j=1;j<=length;j++) {
	  if (abs(tempConfig[j][1]) > radius)
	    radius=abs(tempConfig[j][1]);
	  if (abs(tempConfig[j][2]) > radius)
	    radius=abs(tempConfig[j][2]);
	  if (abs(tempConfig[j][3]) > radius)
	    radius=abs(tempConfig[j][3]);
	}
      }
      
      radius+=1;
      
      printf("\nRadius = %d\n\n",radius);
     
      free_imatrix(tempConfig,0,MAX_LSUBM+1,0,5);
 
    }
  }
}


void moveAlongHome(void)
{

  int count;
  char dummy;

  /* 
     purpose of this routine is to scan along the file until
     an `=' sign is seen 
  */
  
  count=0;
  do {

    fscanf(script,"%1s",&dummy);
    //    cout << dummy;
    count++;

    if (dummy == '='){
      break;
    }

  } while (dummy != '=' && count < 100);

  if (count >=100) {
    printf("Error in reading CCM script: can't find delimiter in time!\n"
	   "Exiting!!\n\n");
    exit(1);
  }

}

int readInInteger(void)
{

  int sum;

  fscanf(script,"%d",&sum);

  return sum;

};
    
double readInFloat(void)
{

  double sum;
  char temp[30];

  fscanf(script,"%29s",temp);
  sum=atof(temp);

  return sum;

};


void setArray(int **typeArray)
{

  int i,j;
    
  for(i=1;i<=13;i++)
    for(j=1;j<=19;j++)
      typeArray[i][j]=0;

  typeArray[1][1]=1;typeArray[1][2]=2;typeArray[1][3]=3;typeArray[1][4]=4;
  typeArray[1][5]=5;

  typeArray[2][1]=6;typeArray[2][2]=7;

  typeArray[3][1]=8;typeArray[3][2]=9;typeArray[3][3]=10;typeArray[3][4]=11;
  typeArray[3][5]=12;typeArray[3][6]=13;typeArray[3][7]=46;
  typeArray[3][8]=47;typeArray[3][9]=48;

  typeArray[4][1]=14;typeArray[4][2]=15;

  typeArray[5][1]=16;typeArray[5][2]=17;typeArray[5][3]=18;typeArray[5][4]=19;
  typeArray[5][5]=20;typeArray[5][6]=21;typeArray[5][7]=49;typeArray[5][8]=50;
  typeArray[5][9]=51;

  typeArray[6][1]=22;typeArray[6][2]=23;typeArray[6][3]=52;

  typeArray[7][1]=24;typeArray[7][2]=25;typeArray[7][3]=53;
  
  typeArray[8][1]=26;

  for(i=27;i<=45;i++)
    typeArray[9][i-26]=i;
  
  typeArray[10][1]=54;

  typeArray[11][1]=55;typeArray[11][2]=56;typeArray[11][3]=57;

  typeArray[12][1]=58;typeArray[12][2]=59;typeArray[12][3]=60;
  typeArray[12][1]=61;

  typeArray[13][1]=62;
}
void definePointsInBT(void)
{

  int i,j,k,l,m,count;

  for(m=0;m<Params.btLength;m++) 
    for (l=0;l<Params.btLength;l++) 
      for(k=0;k<Params.btLength;k++) 
	BTarray[k][l][m]=-99;
  
  for(m=0;m<MAX_BT;m++) {
    lookUpTable[m][1]=0;
    lookUpTable[m][2]=0;
    lookUpTable[m][3]=0;
  }

  if (Params.dim == 1) {
    for(m=0;m<Params.btLength;m++) {
      lookUpTable[m][1]=m;
      lookUpTable[m][2]=0;
      lookUpTable[m][3]=0;
    }
  } else if (Params.dim == 2) {
    count=0;
    for (l=0;l<Params.btLength;l++) {
      for(k=0;k<Params.btLength-l;k++) {
	count++;
	BTarray[k][l][0]=count;
	lookUpTable[count][1]=k;
	lookUpTable[count][2]=l;
	lookUpTable[count][3]=0;
      }
    }
  } else if (Params.dim == 3) {
    count=0;
    for(m=0;m<Params.btLength;m++) {
      for (l=0;l<Params.btLength-m;l++) {
	for(k=0;k<Params.btLength-l;k++) {
	  count++;
	  BTarray[k][l][m]=count;
	  lookUpTable[count][1]=k;
	  lookUpTable[count][2]=l;
	  lookUpTable[count][3]=m;
	}
      }
    }
  }
}  

void defineGeneralisedPositions(void)
{
  
  double tempx, tempy, tempz;
  int i,j,k,l,pointer;
  int temp_x_index;
  int temp_y_index;
  int temp_z_index;
  int temp_cell_index;
  int switchFlag;
  int sorted;
  double tol=0.01;

  double proj_x, proj_y, proj_z;

  // x_array[]; y_array[]; z_array;
  // x_index[]; y_index[]; z_index[]; cell_index[];

  // cells must be ordered in terms of magnitude x and y values in Input.C
  // y values have precedence. (Same with Z of course) 
  

  pointer=0;

  if (Params.dim == 1) {

    for(i=-2*Params.btLength;i<=2*Params.btLength;i++) {
      for(l=1;l<=UnitCell.number;l++) {

	tempx=i*BravaisLattice.a[1]+UnitCell.x[l];
	tempy=i*BravaisLattice.a[2]+UnitCell.y[l];
	tempz=i*BravaisLattice.a[3]+UnitCell.z[l];

	x_array[pointer]=tempx;
	y_array[pointer]=tempy;
	z_array[pointer]=tempz;
	
	x_index[pointer]=i;
	y_index[pointer]=0;
	z_index[pointer]=0;
	cell_index[pointer]=l;

	pointer++;
      }
    }

  } else if (Params.dim==2){

    pointer=1;
    for(j=-2*Params.btLength;j<=2*Params.btLength;j++) {
      for(i=-2*Params.btLength;i<=2*Params.btLength;i++) {
	for(l=1;l<=UnitCell.number;l++) {
	  
	  tempx=i*BravaisLattice.a[1]+j*BravaisLattice.b[1]+UnitCell.x[l];
	  tempy=i*BravaisLattice.a[2]+j*BravaisLattice.b[2]+UnitCell.y[l];
	  
	  x_array[pointer]=tempx;
	  y_array[pointer]=tempy;
	  z_array[pointer]=0;
	  
	  /*      cout << "y = " << y_array[pointer] << " : x = " 
		  << x_array[pointer] 
		  << " : " << pointer << endl;
	  */
	  
	  x_index[pointer]=i;
	  y_index[pointer]=j;
	  z_index[pointer]=0;
	  cell_index[pointer]=l;
	  
	  pointer++;
	}
      } 
    }
    
  } else {
    pointer=1;
    for(k=-2*Params.btLength;k<=2*Params.btLength;k++) {
      for(j=-2*Params.btLength;j<=2*Params.btLength;j++) {
	for(i=-2*Params.btLength;i<=2*Params.btLength;i++) {
	  for(l=1;l<=UnitCell.number;l++) {
	    
	    tempx=i*BravaisLattice.a[1]+j*BravaisLattice.b[1]
	      +k*BravaisLattice.c[1]+UnitCell.x[l];
	    tempy=i*BravaisLattice.a[2]+j*BravaisLattice.b[2]
	      +k*BravaisLattice.c[2]+UnitCell.y[l];
	    tempz=i*BravaisLattice.a[3]+j*BravaisLattice.b[3]
	      +k*BravaisLattice.c[3]+UnitCell.z[l];
	    
	    x_array[pointer]=tempx;
	    y_array[pointer]=tempy;
	    z_array[pointer]=tempz;
	    
	    /*      cout << "y = " << y_array[pointer] << " : x = " 
		    << x_array[pointer] 
		    << " : " << pointer << endl;
	    */
	    
	    x_index[pointer]=i;
	    y_index[pointer]=j;
	    z_index[pointer]=k;
	    cell_index[pointer]=l;
	    
	    pointer++;
	  }
	} 
      }
    }
  }
  
  totalPoints=pointer-1;
  
  // now do a bubble sort.
  
  for(i=1;i<=totalPoints;i++) {
    sorted=1;
    for(j=1;j<totalPoints;j++) {
      
      switchFlag=0;
      
      if (z_array[j+1] < z_array[j]) {
        switchFlag=1;
      } else if (fabs(z_array[j+1] - z_array[j]) < tol) {
	if (y_array[j+1] < y_array[j]) {
          switchFlag=1;
        } else if ((fabs(y_array[j+1] - y_array[j]) < tol) &&
		   (x_array[j+1] < x_array[j])) {
	  switchFlag=1;
	}
      } 
      
      if (switchFlag) {
        
        sorted=0;
	
        tempx=x_array[j+1];
        tempy=y_array[j+1];
        tempz=z_array[j+1];
                
        temp_x_index=x_index[j+1];
        temp_y_index=y_index[j+1];
        temp_z_index=z_index[j+1];
        temp_cell_index=cell_index[j+1];

        x_array[j+1]=x_array[j];
        y_array[j+1]=y_array[j];
        z_array[j+1]=z_array[j];

        x_index[j+1]=x_index[j];
        y_index[j+1]=y_index[j];
        z_index[j+1]=z_index[j];
        cell_index[j+1]=cell_index[j];

        x_array[j]=tempx;
        y_array[j]=tempy;
        z_array[j]=tempz;

        x_index[j]=temp_x_index;
        y_index[j]=temp_y_index;
        z_index[j]=temp_z_index;
        cell_index[j]=temp_cell_index;


      }
    }
    if (sorted)
      break;
  }

  // now check that all of the points are correctly ordered.

  for(i=1;i<totalPoints;i++) {
    //            printf("x=%6.4f y=%6.4f z=%6.4f i=%d\n",
    //	   x_array[i],y_array[i],z_array[i],i);
    
    if (z_array[i+1] < z_array[i]) {
      cout << "Error in  defineGeneralisedPositions. "
	   << "z_array[i+1] < z_array[i]..."
	   << "\nExiting!!\n\n";
      exit(1);
    } else if (fabs(z_array[i+1] - z_array[i]) < tol) {
      if (y_array[i+1] < y_array[i]) {
	cout << "Error in  defineGeneralisedPositions. "
	     << "y_array[i+1] < y_array[i]..."
	     << "\nExiting!!\n\n";
	exit(1);
      } else if ((fabs(y_array[i+1] - y_array[i]) < tol) &&
		 (x_array[i+1] < x_array[i])) {
	cout << "Error in  defineGeneralisedPositions. "
	     << "x_array[i+1] < x_array[i]..."
	     << "\nExiting!!\n\n";
	exit(1);
      }
    } 
  }
}


void convert_numeral_to_char(int number, char *character)
{
  
  int temp;
  int temp1, temp2, temp3,temp4;
  char numerals[11]="0123456789";
  char end[]="\0",null[]="";
  int  i;
  
  for(i=0;i<=9;i++)
    character[i]=null[0];

  if (number >=10000 || number < 0) {
    printf("number either too big or < 0...Exiting\n\n");
    exit(1);
  } else if (number >= 1000 ){
    temp1=number/1000;
    character[0]=numerals[temp1];
    temp2=(number-temp1*1000)/100;
    character[1]=numerals[temp2];
    temp3=(number-temp1*1000-temp2*100)/10;
    character[2]=numerals[temp3];
    temp4=number-temp1*1000-temp2*100-temp3*10;
    character[3]=numerals[temp4];
    character[4]=end[0];
  } else if (number >= 100 ){
    temp1=number/100;
    character[0]=numerals[temp1];
    temp2=(number-temp1*100)/10;
    character[1]=numerals[temp2];
    temp3=number-temp1*100-temp2*10;
    character[2]=numerals[temp3];
    character[3]=end[0];
  } else if (number >= 10 ){
    temp2=number/10;
    character[0]=numerals[temp2];
    temp3=number-temp2*10;
    character[1]=numerals[temp3];
    character[2]=end[0];
  } else {
    character[0]=numerals[number];
    character[1]=end[0];
  }
}

